from datasets import Dataset

from .base import Pipeline

class TabularTrainingPipeline(Pipeline):
    def __call__(
        self,
        data: object,
    ):
        model = self.adapter.train_model(
            data=data,
            **self.kwargs,
        )
        return model




from .base import Pipeline

class TabularEvaluationPipeline(Pipeline):
    def __call__(
        self,
        original_data: object,
        generated_data: object,
    ):
        return self.adapter.evaluate_generated(
            original_data=original_data,
            generated_data=generated_data,
            **self.kwargs,
        )



from huggingface_hub import snapshot_download
from pathlib import Path

from .base import Model
from .synthcity import SynthCityModel
from .synthpop import SynthPopModel

FILE_TO_CLASS = {
    "synthcity.pickle": SynthCityModel,
    "synthpop.pickle": SynthPopModel,
}

MODEL_TO_ADAPTER = {
    SynthCityModel: "SynthCityAdapter",
    SynthPopModel: "SynthPopAdapter",
}

class AutoModel(Model):
    def __init__(self):
        super(AutoModel, self).__init__(saveable=False)
    def from_pretrained(name):
        path = Path(name)
        if not path.exists():
            snapshot_download(repo_id=name)
        if not path.is_dir():
            raise RuntimeError(f"invalid model directory {name} - expected to find a directory {name}")
        for file_name, model_class in FILE_TO_CLASS.items():
            if (path / file_name).is_file():
                return model_class.from_pretrained(name)
        raise RuntimeError(f"could not determine model type for {name} - expected on of: {', '.join(FILE_TO_CLASS.keys())}")



from .base import Pipeline

class TabularSynthesisPipeline(Pipeline):
    def __call__(
        self,
        data: object,
        count: int = None,
    ):
        if count is None:
            # assumption that all formats implement __len__
            count = len(data)
        model = self.adapter.train_model(
            data=data,
            **self.kwargs,
        )
        output = self.adapter.generate_data(
            count=count,
            model=model,
            **self.kwargs,
        )
        if self.output_format is "auto":
            output = self.ensure_output_format(output, output_format=type(data))
        elif self.output_format is not None:
            output = self.ensure_output_format(output)
        return output

class TabularSynthesisDPPipeline(TabularSynthesisPipeline):
    pass


from ..utils.formats import ensure_format

class Adapter:
    def __init__(self, input_formats):
        self.input_formats = input_formats
    def ensure_input_format(self, data, **kwargs):
        return ensure_format(data, self.input_formats, **kwargs)



from ..adapters import *
from .base import Pipeline
from ..models.auto import MODEL_TO_ADAPTER

class TabularGenerationPipeline(Pipeline):
    def __call__(
        self,
        count: int = 1,
    ):
        if self.adapter is None:
            self.adapter = eval(MODEL_TO_ADAPTER[self.kwargs["model"].__class__])()
        output = self.adapter.generate_data(
            count=count,
            **self.kwargs,
        )
        if self.output_format is not None:
            output = self.ensure_output_format(output)
        return output



from typing import Any, Dict, Optional, Tuple, Type, Union

from ..adapters import (
    SynthCityAdapter,
)
from ..adapters.base import (
    Adapter,
)
from .base import (
    PipelineRegistry,
)
from .tabular_evaluation import TabularEvaluationPipeline
from .tabular_generation import TabularGenerationPipeline
from .tabular_synthesis import TabularSynthesisPipeline, TabularSynthesisDPPipeline
from .tabular_training import TabularTrainingPipeline

# Register all the supported tasks here
TASK_ALIASES = {
    "evaluate": "tabular-evaluation",
    "generate": "tabular-generation",
    "synthesize": "tabular-synthesis",
    "train": "tabular-training",
}
SUPPORTED_TASKS = {
    "tabular-evaluation": {
        "impl": TabularEvaluationPipeline,
        "adapter": SynthCityAdapter(),
        "type": "tabular",
    },
    "tabular-generation": {
        "impl": TabularGenerationPipeline,
        "adapter": None,
        "type": "tabular",
    },
    "tabular-synthesis": {
        "impl": TabularSynthesisPipeline,
        "adapter": SynthCityAdapter(),
        "type": "tabular",
    },
    # This task is a special case as it's parametrized by EPSILON and DELTA.
    "tabular-synthesis-dp": {
        "impl": TabularSynthesisDPPipeline,
        "adapter": SynthCityAdapter(plugin="dpgan"),
        "type": "tabular",
    },
    "tabular-training": {
        "impl": TabularTrainingPipeline,
        "adapter": SynthCityAdapter(),
        "type": "tabular",
    },
}

PIPELINE_REGISTRY = PipelineRegistry(supported_tasks=SUPPORTED_TASKS, task_aliases=TASK_ALIASES)

def check_task(task: str) -> Tuple[str, Dict, Any]:
    """
    Checks an incoming task string, to validate it's correct and return the default Pipeline and Model classes, and
    default models if they exist.

    Args:
        task (`str`):
            The task defining which pipeline will be returned. Currently accepted tasks are:

            - `"tabular-synthesis"`
            - `"tabular-synthesis_EPSILON_DELTA"`

    Returns:
        (normalized_task: `str`, task_defaults: `dict`, task_options: (`tuple`, None)) The normalized task name
        (removed alias and options). The actual dictionary required to initialize the pipeline and some extra task
        options for parametrized tasks like "tabular-synthesis_EPSILON_DELTA"


    """
    return PIPELINE_REGISTRY.check_task(task)

def pipeline(
    task: str = None,
    adapter: Optional[Union[Adapter, str]] = None,
    pipeline_class: Optional[Type] = None,
    **kwargs,
):
    normalized_task, targeted_task, task_options = check_task(task)
    if pipeline_class is None:
        pipeline_class = targeted_task["impl"]
    if adapter is None:
        adapter = targeted_task["adapter"]

    return pipeline_class(task=task, adapter=adapter, **kwargs)




from datasets import Dataset
from numpy import ndarray, array
from pandas import DataFrame
from synthcity.plugins.core.dataloader import GenericDataLoader

SUPPORTED_FORMATS = [
    list,
    ndarray,
    DataFrame,
    Dataset,
    GenericDataLoader,
]

def ensure_format(data, target_formats, **kwargs):
    source_format = type(data)
    if source_format not in SUPPORTED_FORMATS:
        raise ValueError(f"unknown source format {source_format}")
    for target_format in target_formats:
        if target_format not in SUPPORTED_FORMATS:
            raise ValueError(f"unknwown target format {target_format} in {target_formats}")
    if source_format in target_formats:
        return data
    for target_format in target_formats:
        if target_format == Dataset:
            if source_format in (list, ndarray):
                data, source_format = DataFrame(data), DataFrame
            if source_format == GenericDataLoader:
                data, source_format = data.dataframe(), DataFrame
            if source_format == DataFrame:
                ds = Dataset.from_pandas(data)
                ds = ds.remove_columns([feature for feature in ds.features if feature.startswith("__") and feature.endswith("__")])
                return ds
            raise ValueError(f"cannot convert {source_format} to {target_format}")
        if target_format == DataFrame:
            if source_format in (list, ndarray):
                return DataFrame(data)
            if source_format == Dataset:
                return data.to_pandas()
            if source_format == GenericDataLoader:
                return data.dataframe()
            raise ValueError(f"cannot convert {source_format} to {target_format}")
        if target_format == GenericDataLoader:
            if source_format == Dataset:
                data, source_format = data.to_pandas(), DataFrame
            if source_format in (list, ndarray, DataFrame):
                return GenericDataLoader(data, **kwargs)
            raise ValueError(f"cannot convert {source_format} to {target_format}")
        if target_format == list:
            if source_format == Dataset:
                data, source_format = data.to_pandas(), DataFrame
            if source_format == GenericDataLoader:
                data, source_format = data.dataframe(), DataFrame
            if source_format == DataFrame:
                data, source_format = data.to_numpy(), ndarray
            if source_format == ndarray:
                return data.tolist()
            raise ValueError(f"cannot convert {source_format} to {target_format}")
        if target_format == ndarray:
            if source_format == Dataset:
                data, source_format = data.to_pandas(), DataFrame
            if source_format == GenericDataLoader:
                data, source_format = data.dataframe(), DataFrame
            if source_format == DataFrame:
                return data.to_numpy()
            if source_format == list:
                return array(data)
            raise ValueError(f"cannot convert {source_format} to {target_format}")
        raise ValueError(f"cannot convert {source_format} to any of {target_formats}")


from pandas import DataFrame
from synthpop import Synthpop

from .base import Adapter
from ..models.synthpop import SynthPopModel

MAP_TYPES = {
    'int64': 'int',
    'int8' : 'int',
    'category': 'category',
}

class SynthPopAdapter(Adapter):
    def __init__(self, input_formats=(DataFrame,)):
        super(SynthPopAdapter, self).__init__(input_formats)
    def train_model(self, data, plugin=None):
        data = self.ensure_input_format(data)
        dtypes = {col:MAP_TYPES[dtype.name] for col, dtype in data.dtypes.to_dict().items()}
        model = Synthpop()
        model.fit(data, dtypes)
        return SynthPopModel(model)
    def generate_data(self, count, model):
        result = model.model.generate(k=count)
        return result
    def evaluate_generated(self, original_data, generated_data):
        raise NotImplementedError("evaluation not implemented for SynthPop")



import logging
import os
import sys
import threading

from typing import Optional

_lock = threading.Lock()
_default_handler: Optional[logging.Handler] = None

log_levels = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

_default_log_level = logging.WARNING

def _get_default_logging_level():
    """
    If TRANSFORMERS_VERBOSITY env var is set to one of the valid choices return that as the new default level. If it is
    not - fall back to `_default_log_level`
    """
    env_level_str = os.getenv("SYNTHESIZERS_VERBOSITY", None)
    if env_level_str:
        if env_level_str in log_levels:
            return log_levels[env_level_str]
        else:
            logging.getLogger().warning(
                f"Unknown option SYNTHESIZERS_VERBOSITY={env_level_str}, "
                f"has to be one of: { ', '.join(log_levels.keys()) }"
            )
    return _default_log_level

def _get_library_name() -> str:
    return __name__.split(".")[0]

def _get_library_root_logger() -> logging.Logger:
    return logging.getLogger(_get_library_name())

def _configure_library_root_logger() -> None:
    global _default_handler

    with _lock:
        if _default_handler:
            # This library has already configured the library root logger.
            return
        _default_handler = logging.StreamHandler()  # Set sys.stderr as stream.
        _default_handler.flush = sys.stderr.flush

        # Apply our default configuration to the library root logger.
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(_default_handler)
        library_root_logger.setLevel(_get_default_logging_level())
        library_root_logger.propagate = False

def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Return a logger with the specified name.

    This function is not supposed to be directly accessed unless you are writing a custom transformers module.
    """

    if name is None:
        name = _get_library_name()

    _configure_library_root_logger()
    return logging.getLogger(name)



from pandas import DataFrame
from syntheval import SynthEval

from .base import Adapter

class SynthEvalAdapter(Adapter):
    def __init__(self, input_formats=(DataFrame,), config: str = 'full', cat_cols: str = None, target_col: str = None):
        super(SynthEvalAdapter, self).__init__(input_formats)
        self.config = config
        self.cat_cols = cat_cols
        self.target_col = target_col
    def train_model(self, data):
        raise NotImplementedError("training not implemented for SynthEval")
    def generate_data(self, count, model):
        raise NotImplementedError("generation not implemented for SynthEval")
    def evaluate_generated(self, original_data, generated_data, config: str = 'full', cat_cols: str = None, target_col: str = None):
        original_data = self.ensure_input_format(original_data)
        generated_data = self.ensure_input_format(generated_data)
        cat_cols = self.cat_cols if cat_cols is None else cat_cols
        evaluator = SynthEval(real=original_data)
        config = self.config if config is None else config
        target_col = self.target_col if target_col is None else target_col
        if config == 'full':
            evaluator.full_eval(synthetic_data=generated_data, target_col=target_col)
        elif config == 'fast':
            evaluator.fast_eval(synthetic_dataframe=generated_data, target=target_col)
        else:
            raise ValueError(f"unknown config {config} -- use either 'full' or 'fast'")
        return evaluator.res_dict



from huggingface_hub import hf_hub_download
from pathlib import Path
from synthcity.plugins import Plugins

from .base import Model

MODEL_FILE = "synthcity.pickle"

class SynthCityModel(Model):
    def __init__(self, model):
        super(SynthCityModel, self).__init__(saveable=True)
        self.model = model
    def save_pretrained(self, name):
        path = Path(name)
        path.mkdir(parents=True, exist_ok=True)
        buff = self.model.save()
        with open(path / MODEL_FILE, "wb") as f:
            f.write(buff)
    def from_pretrained(name):
        path = Path(name)
        if not path.exists():
            hf_hub_download(repo_id=name, filename=MODEL_FILE)
        if not path.is_dir() or not (path / MODEL_FILE).is_file():
            raise RuntimeError(f"invalid synthcity model directory {name} - expected to find a file {name / MODEL_FILE}")
        with open(path / MODEL_FILE, "rb") as f:
            buff = f.read()
        model = Plugins().load(buff)
        return SynthCityModel(model)
    def __repr__(self):
        return f"SynthCityModel({repr(self.model)})"



from .synthcity import SynthCityAdapter
from .syntheval import SynthEvalAdapter
from .synthpop import SynthPopAdapter

NAME_TO_ADAPTER = {
    "synthcity": SynthCityAdapter,
    "synthpop": SynthPopAdapter,
    "syntheval": SynthEvalAdapter,
}


from synthcity.metrics.eval_statistical import AlphaPrecision
from synthcity.plugins.core.dataloader import GenericDataLoader
from synthcity.plugins import Plugins

from .base import Adapter
from ..models.synthcity import SynthCityModel

class SynthCityAdapter(Adapter):
    def __init__(self, plugin = 'adsgan', evaluator_class = AlphaPrecision, input_formats=(GenericDataLoader,)):
        super(SynthCityAdapter, self).__init__(input_formats)
        self.plugin = plugin
        self.evaluator_class = evaluator_class
    def train_model(self, data, plugin=None):
        data = self.ensure_input_format(data)
        model = Plugins().get(self.plugin if plugin is None else plugin)
        model.fit(data)
        return SynthCityModel(model)
    def generate_data(self, count, model):
        result = model.model.generate(count=count)
        return result
    def evaluate_generated(self, original_data, generated_data, evaluator_class=None):
        original_data = self.ensure_input_format(original_data)
        generated_data = self.ensure_input_format(generated_data)
        evaluator_class = self.evaluator_class if evaluator_class is None else evaluator_class
        evaluator = evaluator_class()
        return evaluator.evaluate(original_data, generated_data)



from typing import Any, Dict, List, Optional, Tuple, Type, Union

from ..adapters import NAME_TO_ADAPTER
from ..adapters.base import Adapter
from ..utils.formats import ensure_format
from ..utils import logging

logger = logging.get_logger(__name__)

class PipelineRegistry:
    def __init__(self, supported_tasks: Dict[str, Any], task_aliases: Dict[str, str]) -> None:
        self.supported_tasks = supported_tasks
        self.task_aliases = task_aliases

    def get_supported_tasks(self) -> List[str]:
        supported_task = list(self.supported_tasks.keys()) + list(self.task_aliases.keys())
        supported_task.sort()
        return supported_task

    def check_task(self, task: str) -> Tuple[str, Dict, Any]:
        if task in self.task_aliases:
            task = self.task_aliases[task]
        if task in self.supported_tasks:
            targeted_task = self.supported_tasks[task]
            return task, targeted_task, None

        if task.startswith("tabular-synthesis-dp"):
            tokens = task.split("_")
            if len(tokens) == 3 and tokens[0] == "tabular-synthesis-dp":
                targeted_task = self.supported_tasks["tabular-synthesis"]
                task = "tabular-synthesis"
                return task, targeted_task, (float(tokens[1]), float(tokens[2]))
            raise KeyError(f"Invalid tabular-synthesis-dp task {task}, use 'tabular-synthesis-dp_EPSILON_DELTA' format with EPSILON and DELTA being floats")

        raise KeyError(
            f"Unknown task {task}, available tasks are {self.get_supported_tasks() + ['tabular-synthesis-dp_EPSILON_DELTA']}"
        )

class Pipeline():
    def __init__(self,
            task: str,
            adapter: Optional[Union[Adapter, str]] = None,
            output_format: Optional[Type] = None,
            **kwargs,
        ):
        self.task = task
        if isinstance(adapter, str):
            adapter = NAME_TO_ADAPTER[adapter]()
        self.adapter = adapter
        self.output_format = output_format
        self.kwargs = kwargs
    def ensure_output_format(self, data, output_format=None, **kwargs):
        return ensure_format(data, (self.output_format if output_format is None else output_format,), **kwargs)
